import asyncio
import datetime

from ._sentinel import MISSING


async def sleep_until(until: datetime.datetime) -> None:
    """
    Sleep until a certain time.

    .. versionadded:: 1.3

    Parameters
    -----------
    until: :class:`datetime.datetime`
        The time to sleep until.

    Raises
    -------
    TypeError
        `until` is not `datetime.datetime.`
    """
    if not isinstance(until, datetime.datetime):  # type: ignore
        raise TypeError(f'Expected datetime.datetime, not {until.__class__.__name__!r}')

    seconds = (until - datetime.datetime.now(tz=until.tzinfo)).total_seconds()
    await asyncio.sleep(seconds)


async def today_sleep_until(hours: int = MISSING, minutes: int = MISSING, seconds: int = MISSING):
    """
    Sleep until a certain time. The difference between this and :func:`sleep_until` is that this only sleeps within the current day.

    .. versionadded:: 1.3

    Parameters
    -----------
    hours: :class:`int`
        The hours in the time.
    minutes: :class:`int`
        The minutes in the time.
    seconds: :class:`int`
        The seconds in the time.


    Raises
    -------
    TypeError
        No arguments were provided.


    Example
    ---------
    .. code:: py

        import asyncify

        async def main():
            await asyncify.today_sleep_until(hours=9, minutes=30)  # sleeps until 9:30 AM
    """
    if not hours or not minutes or not seconds:
        raise TypeError('You must provide hours, minutes, or seconds.')

    now = datetime.datetime.now()
    total_seconds_now = (now.hour * 3600) + (now.minute * 60) + now.second
    total_seconds = (hours * 3600) + (minutes * 60) + (seconds * 1)

    sleep_time = total_seconds - total_seconds_now
    await asyncio.sleep(sleep_time)



from __future__ import annotations

import inspect
import re
from typing import TYPE_CHECKING, Any, Callable, Generic, Optional, TypeVar, Union

if TYPE_CHECKING:
    from types import FrameType
    from typing_extensions import Self
    from ._types import Coro


__all__ = ('HybridFunction', 'hybrid_function')


T_sync = TypeVar('T_sync')
T_async = TypeVar('T_async')


class HybridFunction(Generic[T_sync, T_async]):
    regex = re.compile(r'await\s+(\w|\.)*\s*\(.*\)')
    name_regex: re.Pattern[str]

    def __init__(
        self,
        name: str,
        sync_callback: Callable[..., T_sync],
        async_callback: Callable[..., Coro[T_async]],
    ):
        self._name = name
        self.sync_callback = sync_callback
        self.async_callback = async_callback

        self._instance: Optional[object] = None

        self.name_regex = re.compile(rf'\(*{self._name}\)*\s*')

    def __repr__(self) -> str:
        return f'HybridFunction({self._name!r}, {self.sync_callback!r}, {self.async_callback!r})'

    @property
    def __name__(self) -> str:
        return self._name

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, self.__class__)
            and other._name == self._name
            and other.sync_callback is self.sync_callback
            and other.async_callback is self.async_callback
        )

    def __get__(self, instance: object, owner: type) -> Self:
        new_self = self.__class__(self._name, self.sync_callback, self.async_callback)
        new_self._instance = instance
        return new_self

    def _check_regex(self, code_context: str) -> bool:
        search = self.regex.search(code_context)
        if not search:
            return False
        return True

    def _get_frame(self, current_frame: FrameType) -> str:
        for frame in inspect.getouterframes(current_frame):
            if not frame.code_context:
                continue
            if self.name_regex.search(frame.code_context[0]):
                return frame.code_context[0]
        raise RuntimeError('Could not tell if it should call sync or async.')

    def __call__(self, *args: Any, **kwargs: Any) -> Union[T_sync, Coro[T_async]]:
        if self._instance:
            args = (self._instance, *args)

        frame = inspect.currentframe()
        assert frame is not None

        code_context = self._get_frame(frame).strip()
        if self._check_regex(code_context):
            return self.async_callback(*args, **kwargs)
        return self.sync_callback(*args, **kwargs)


def hybrid_function(
    name: str, sync_callback: Callable[..., T_sync], async_callback: Callable[..., Coro[T_async]]
) -> HybridFunction[T_sync, T_async]:
    """
    Do multiple things depending on whether it was awaited or not!


    Parameters
    ----------
    name: :class:`str`
        The name of the new function. This must be the same as the function name to work.
    sync_callback: ``Callable[..., Any]``
        The callable to call if it is not awaited.
    async_callback: ``Callable[..., Coroutine]``
        The callable to call if it is awaited.


    Example
    --------
    .. code:: py

        import asyncify
        import discord  # discord.py example

        class Client(discord.Client):
            get_or_fetch_user = asyncify.hybrid_function(
                                'get_or_fetch_user',
                                 discord.Client.get_user,
                                 discord.Client.fetch_user
            )

        client = Client()

        client.get_or_fetch_user(739510612652195850)  # sync cache lookup
        await client.get_or_fetch_user(739510612652195850)  # async api call


    .. warning::
        Make the to name the function uniquely. Functions with the same name could be called unexpectedly.
    """
    return HybridFunction[T_sync, T_async](name, sync_callback, async_callback)



from typing import Any


class MissingSentinel:
    def __repr__(self):
        return '...'

    def __mul__(self, other: Any):
        return 0

    def __eq__(self, other: Any):
        return False


MISSING: Any = MissingSentinel()



from __future__ import annotations

import asyncio
import functools
import sys
from typing import TYPE_CHECKING, Any, Callable, Dict, Set, Tuple, Type, TypeVar

if TYPE_CHECKING:
    from typing_extensions import ParamSpec


__all__ = ('EventsEventLoopPolicy',)


T = TypeVar('T')

if TYPE_CHECKING:
    P = ParamSpec('P')
else:
    P = TypeVar('P')


_UNIX_ONLY_NAMES: Tuple[str, ...] = ('get_child_watcher', 'set_child_watcher')

_VALID_NAMES: Tuple[str, ...] = ('get_event_loop', 'set_event_loop', 'new_event_loop', *_UNIX_ONLY_NAMES)


class EventsEventLoopPolicy(asyncio.DefaultEventLoopPolicy):
    """
    Call a functions whenever certain things happen in asyncio!
    This is done using the event loop policy.

    Example
    ---------
    .. code:: py

        import asyncio
        import asyncify

        policy = asyncify.EventsEventLoopPolicy()

        @policy.event
        def new_event_loop():
            print('New event loop being created.')

        asyncio.set_event_loop_policy(policy)


        async def main():
            ...

        asyncio.run(main())  # prints 'New event loop being created.'
        # asyncio.run creates an event loop with `new_event_loop`

    .. warning::
        `asyncio.get_event_loop` won't call its event if there is a running and set event loop.

    .. versionadded:: 1.1
    """

    _OLDS: Dict[str, Callable[..., Any]] = {}

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._set_olds()
        self._registered_events: Set[str] = set()

    @classmethod
    def _set_olds(cls) -> None:
        cls._OLDS['get_event_loop'] = cls.get_event_loop
        cls._OLDS['set_event_loop'] = cls.set_event_loop
        cls._OLDS['new_event_loop'] = cls.new_event_loop

        if sys.platform != 'win32':
            cls._OLDS['get_child_watcher'] = cls.get_child_watcher
            cls._OLDS['set_child_watcher'] = cls.set_child_watcher

    def __repr__(self) -> str:
        return (
            f'<EventsEventLoopPolicy base={self.__class__.__bases__[0]!r},'
            f' registered_events={self._registered_events!r}>'
        )

    def event(self, func: Callable[P, T]) -> Callable[P, T]:
        """|deco|

        Register a function to be called when an event loop policy method is called.
        The name of the functions should match the event loop policy method.
        The valid names are ``get_event_loop``, ``set_event_loop``, ``new_event_loop``,
        ``get_child_watcher``, ``set_child_watcher``

        Example
        ---------
        .. code:: py

            @policy.event
            def new_event_loop():
                print('New event loop being created.')

        .. note::
            Using it multiple times on the same method will overwrite the old one.

        Raises
        -------
        TypeError
            `func` is not a callable.
        RuntimeError
            The name either is invalid or is not supported on Windows.
        """
        if not isinstance(func, Callable):
            raise TypeError(f'Expected a callable, got {func.__class__.__name__!r}')

        if sys.platform == 'win32' and func.__name__ in _UNIX_ONLY_NAMES:
            raise RuntimeError(f'{func.__name__!r} is not supported on windows.')

        if func.__name__ not in _VALID_NAMES:
            raise RuntimeError(f'{func.__name__!r} is not a valid function name. {_VALID_NAMES} are valid.')

        old: Callable[..., Any] = getattr(super(), func.__name__)

        @functools.wraps(old)
        def updated(*args: P.args, **kwargs: P.kwargs) -> Any:
            func(*args, **kwargs)
            return old(*args, **kwargs)

        setattr(self, old.__name__, updated)
        self._registered_events.add(old.__name__)

        return func

    def remove_event(self, name: str) -> None:
        """
        Remove an event from the callbacks that would be called.

        Parameters
        ----------
        name: class:`str`
            The name of the event.

        Raises
        -------
        RuntimeError
            `name` is not a registered event.
        """
        if name not in self._registered_events:
            raise RuntimeError(f'{name!r} is not a registered event.')

        setattr(self, name, self._OLDS[name])
        self._registered_events.remove(name)

    @classmethod
    def change_base_policy(cls, policy_cls: Type[asyncio.AbstractEventLoopPolicy]) -> None:
        """
        The actual things the event loop policy does are based on your platform.
        The default base loop policy is
        `asyncio.DefaultLoopPolicy <https://docs.python.org/3/library/asyncio-policy.html#asyncio.DefaultEventLoopPolicy>`_
        but can be changed.

        Parameters
        ------------
        policy_cls: Type[:class:`asyncio.AbstractEventLoopPolicy`]
            This class must inherit from
            `asyncio.AbstractEventLoopPolicy <https://docs.python.org/3/library/asyncio-policy.html#asyncio.AbstractEventLoopPolicy>`_


        This example uses `uvloop`.

        Example
        --------
        .. code:: py

            import asyncio
            import asyncify
            import uvloop

            asyncify.EventsEventLoopPolicy.change_base_policy(uvloop.EventLoopPolicy)

            policy = asyncify.EventsEventLoopPolicy()

            @policy.event
            def new_event_loop():
                ...

            loop = asyncio.new_event_loop()
            print(loop)  # <uvloop.Loop running=False closed=False debug=False>

        Raises
        -------
        TypeError
            `policy_cls` does not inherit from `asyncio.AbstractEventLoopPolicy`.
        """
        if asyncio.AbstractEventLoop not in policy_cls.__bases__:
            raise TypeError('policy_cls must inherited from asyncio.AbstractEventLoopPolicy.')

        cls.__bases__: Tuple[Type[Any], ...] = (policy_cls, *cls.__bases__[1:])
        cls._set_olds()



from __future__ import annotations

import asyncio
import inspect
import functools

from typing import TYPE_CHECKING, Any, Callable, Coroutine, TypeVar

if TYPE_CHECKING:
    from typing_extensions import ParamSpec


__all__ = ('asyncify_func', 'syncify_func')


T = TypeVar('T')

if TYPE_CHECKING:
    P = ParamSpec('P')
else:
    P = TypeVar('P')


def asyncify_func(func: Callable[P, T]) -> Callable[P, Coroutine[Any, Any, T]]:
    """|deco|

    Make a synchronous function into an asynchronous function by running it in a separate thread.

    Example
    --------
    .. code:: py

        import asyncify
        import requests

        @asyncify.asyncify_func
        def get(url):
            return requests.get(url).text

        # `get` is no longer a blocking function
        # it is now a coroutine function

        async def main():
            text = await get('https://python.org')

        # this is very useful to turn a blocking library into an async library
        get = asyncify.asyncify_func(requests.get)

    .. note::
        This function uses the default loop executor.
        Change it with `loop.set_default_executor <https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.set_default_executor>`_.

    Raises
    -------
    TypeError
        The object passed in was not a function.
    """
    if not inspect.isfunction(func):
        raise TypeError(f'Expected a callable function, got {func.__class__.__name__!r}')

    @functools.wraps(func)
    async def async_func(*args: P.args, **kwargs: P.kwargs) -> T:
        new_func = functools.partial(func, *args, **kwargs)

        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, new_func)

    return async_func


def syncify_func(func: Callable[P, Coroutine[Any, Any, T]]) -> Callable[P, T]:
    """|deco|

    Make an asynchronous function a synchronous function.

    Example
    --------
    .. code:: py

        import asyncio
        import asyncify

        @asyncify.syncify_func
        async def coroutine_func():
            await asyncio.sleep(5)
            print('Done')

        coroutine_func()  # can be directly called


    .. note::

        This is equivalent to the following:

        .. code:: py

            loop = asyncio.get_event_loop()
            loop.run_until_complete(coroutine_func())

    .. warning::
        There must be a running event loop.

    Raises
    -------
    TypeError
        The object passed was not a coroutine function.
    """
    if not inspect.isfunction(func):
        raise TypeError(f'Expected a callable function, got {func.__class__.__name__!r}')

    @functools.wraps(func)
    def sync_func(*args: P.args, **kwargs: P.kwargs) -> T:
        loop = asyncio.get_running_loop()
        return loop.run_until_complete(func(*args, **kwargs))

    return sync_func



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from ._types import Coro


__all__ = ('run',)


T = TypeVar('T')


# remove in v2
def run(main: Coro[T], *, debug: bool = False) -> T:
    import warnings

    warnings.warn('asyncify.run has been deprecated since 1.3. Use asyncio.run instead.', category=DeprecationWarning)
    return asyncio.run(main, debug=debug)



from typing import Any, Awaitable, Coroutine, Callable, Type, TypeVar


T = TypeVar('T')

Coro = Coroutine[Any, Any, T]

CallableT = TypeVar('CallableT', bound=Callable[..., Any])
TypeT = TypeVar('TypeT', bound=Type[Any])

NoArgAwaitable = Callable[[], Awaitable[T]]



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    NoReturn,
    TypeVar,
    Optional,
)

if TYPE_CHECKING:
    from typing_extensions import Self
    from ._types import NoArgAwaitable


__all__ = ('AsyncIterable', 'async_iter')


T = TypeVar('T')


class AsyncIterable(Generic[T]):
    __slots__ = ('iterable', 'iterator', 'before', 'after')

    def __init__(
        self,
        iterable: Iterable[T],
        *,
        before: Optional[NoArgAwaitable[Any]] = None,
        after: Optional[NoArgAwaitable[Any]] = None,
    ):
        self.iterable = iterable
        self.iterator: Optional[Iterator[T]] = None

        self.before = before
        self.after = after

    def __repr__(self) -> str:
        return f'AsyncIterable({self.iterator!r}, before={self.before!r}, after={self.after!r})'

    def __await__(self) -> Generator[Any, Any, List[T]]:
        return self.flatten().__await__()

    def __iter__(self) -> NoReturn:
        raise TypeError(f'{self.__class__.__name__!r} object is not iterable, use async for instead.')

    def __aiter__(self) -> "Self":
        self.iterator = iter(self.iterable)
        return self

    async def __anext__(self) -> T:
        assert self.iterator is not None

        if self.before is not None:
            await self.before()

        try:
            item = next(self.iterator)
        except StopIteration:
            self.iterator = None
            raise StopAsyncIteration

        if self.after is not None:
            await self.after()

        return item

    async def flatten(self) -> List[T]:
        return [item async for item in self]


def async_iter(
    iterable: Iterable[T],
    before: Optional[NoArgAwaitable[Any]] = None,
    after: Optional[NoArgAwaitable[Any]] = None,
) -> AsyncIterable[T]:
    """
    Asynchronously iterate through an iterable while calling an async callback before and/or after each iteration.

    Parameters
    ------------
    iterable: :class:`Iterable`
        The iterable to iterator over.
    before: Optional[``Callable[<no_parameters>, Awaitable[Any]]``]
        An optional asynchronous callable for before the iteration.
    after: Optional[``Callable[<no_parameters>, Awaitable[Any]]``]
        An optional asynchronous callable for after the iteration.


    .. note::
        `before` and `after` must not take any parameters.

    .. warning::
        `before` will be called once more than `after` due to not knowing when to stop iterating until after
        `before` is called.

    Example
    ---------
    .. code:: py

        import asyncio
        import functools
        import asyncify

        async def main():
            sleep = functools.partial(asyncio.sleep, 1)

            async for number in async_iter([1, 2, 3], sleep, before=True):
                print(f'{number} seconds have passed.')

    Raises
    -------
    TypeError
        The object passed was not an iterable.
    """
    if not hasattr(iterable, '__iter__'):
        raise TypeError(f'Expected iterable object, got {iterable.__class__.__name__!r}')

    return AsyncIterable[T](iterable, before=before, after=after)



# type: ignore

from .cls import *
from .func import *
from .events import *
from .hybrid import *
from .iterable import *
from .runners import *
from .sleep import *


__all__ = cls.__all__ + func.__all__ + events.__all__ + hybrid.__all__ + iterable.__all__ + runners.__all__


__title__ = 'asyncify'
__author__ = 'The Master'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022-present The Master'
__version__ = '1.4.3'



import re
from setuptools import setup


with open('requirements.txt') as f:
    requirements = f.read().splitlines()


with open('asyncify/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)


with open('README.rst', 'r') as rm:
    readme = rm.read()


# fmt: off
packages = [
    'asyncify'
    ]


extras_require = {
    'docs': [
        'sphinx-press-theme'
        ]
    }
# fmt: on

setup(
    name='asyncify-python',
    author='The Master',
    version=version,
    packages=packages,
    license='MIT',
    description='A python library to make things async!',
    project_urls={
        'GitHub': 'https://github.com/chawkk6404/asyncify',
        'Documentation': 'https://asyncify.readthedocs.io/en/latest',
    },
    long_description=readme,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    python_requires='>=3.7.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Type

from .func import asyncify_func

if TYPE_CHECKING:
    from ._types import CallableT, TypeT


__all__ = ('asyncify_class', 'ignore')


def ignore(func: CallableT) -> CallableT:
    """|deco|

    Ignore a function in a class when using :func:`asyncify.asyncify_class`.
    """
    func._asyncify_ignore = True  # type: ignore # we are assigning new attribute here
    return func


def asyncify_class(cls: TypeT) -> TypeT:
    """|deco|

    Turn a classes methods into async functions.
    This uses :func:`asyncify.asyncify_func`.
    This ignores methods marked with :func:`asyncify.ignore` and `dunder` methods.

    Example
    ---------
    .. code:: py

        import asyncify
        import requests

        @asyncify.asyncify_class
        class RequestsClient:
            def __init__(self):  # ignored by asyncify
                self.session = requests.Session()

            def request(self, method, url):  # now a coroutine function
                return self.session.request(method, url)

        # can also be used like this
        RequestsClient = asyncify.asyncify_class(requests.Session)

        client = RequestsClient()

        async def main():
            await client.request('GET', 'https://python.org')

    Raises
    -------
    TypeError
        The object passed was not a class.
    """
    if not isinstance(cls, Type):
        raise TypeError(f'Expected class, got {cls.__class__.__name__!r}')

    for name, func in inspect.getmembers(cls):
        if not inspect.isfunction(func) or getattr(func, '_asyncify_ignore', False) or name.startswith('__'):
            continue

        func = asyncify_func(func)
        setattr(cls, name, func)

    return cls
